/*
 * Copyright (c) 2003-2005 Erez Zadok
 * Copyright (c) 2003-2005 Charles P. Wright
 * Copyright (c) 2003-2005 Mohammad Nayyer Zubair
 * Copyright (c) 2003-2005 Puja Gupta
 * Copyright (c) 2003-2005 Harikesavan Krishnan
 * Copyright (c) 2003-2005 Stony Brook University
 * Copyright (c) 2003-2005 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */
/*
 *  $Id: lookup.c,v 1.2 2005/05/05 12:28:15 cwright Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "unionfs.h"

static int is_opaque_dir(struct dentry *dentry, int bindex);

struct dentry *unionfs_lookup_backend(struct inode *parent, 
				struct dentry *dentry, int interpose_flag)
{
	int err = 0;
	struct dentry *hidden_dentry = NULL;
	struct dentry *wh_hidden_dentry = NULL;
	struct dentry *hidden_dir_dentry = NULL;
	struct dentry *parent_dentry;
	int bindex, bstart, bend;
	int dentry_count;
	int is_first_negative = 0;
	int first_dentry_offset = -1;
	struct dentry *first_hidden_dentry;
	char *name = NULL;

	print_entry_location();
	PASSERT(dentry);
	/* No dentries should get created for possible whiteout names. */
	if (!strncmp(dentry->d_name.name, ".wh.", 4)) {
		err = -EPERM;
		goto out;
	}
	if (!strncmp(dentry->d_name.name, UNIONFS_DIR_OPAQUE_NAME,
				sizeof(UNIONFS_DIR_OPAQUE_NAME) - 1)) {
		err = -EPERM;
		goto out;
	}

	parent_dentry = dentry->d_parent;
	PASSERT(parent_dentry);
	bstart = dbstart(parent_dentry);
	bend = dbend(parent_dentry);
	ASSERT(bstart >= 0);

	fist_print_dentry("IN unionfs_lookup:", dentry);
	fist_print_inode("IN unionfs_lookup:", parent);

	/* must initialize dentry operations */
	dentry->d_op = &unionfs_dops;

	/* allocate space for hidden dentries */
	if (!dtopd_lhs(dentry)) {
		dtopd_lhs(dentry) =  (struct unionfs_dentry_info *) 
				KMALLOC(sizeof(struct unionfs_dentry_info), 
				GFP_UNIONFS);
		if (!dtopd_lhs(dentry)) {
			err = -ENOMEM;
			goto out;
		}
		init_dtopd(dentry, parent->i_sb);
	} else {
		reinit_dtopd(dentry, parent->i_sb);
	}
	dtohd_ptr(dentry) = KMALLOC(sizeof(struct dentry *) *
				sbmax(parent->i_sb), GFP_UNIONFS);
	if (!dtohd_ptr(dentry)) {
		err = -ENOMEM;
		goto out_free_dtopd;
	}
	init_dtohd_ptr(dentry, sbmax(parent->i_sb));

	/* these fields are useful when all dentries are negative */
	dentry_count = 0;      /* the number of positive dentries found */
	is_first_negative = 0;   /* boolean: is the first dentry negative? */
	first_hidden_dentry = NULL;  /* pointer to the first dentry */

	ASSERT(bstart >= 0);
	for (bindex = bstart; bindex <= bend; bindex++) {
		hidden_dir_dentry = dtohd_index(parent_dentry, bindex);

		/* if the parent hidden dentry does not exist skip this */
		if (!(hidden_dir_dentry && hidden_dir_dentry->d_inode))
			continue;

		PASSERT(hidden_dir_dentry->d_inode);
		/* also skip it if the parent isn't a directory. */
		if (!S_ISDIR(hidden_dir_dentry->d_inode->i_mode))
			continue;

		/* Reuse the whiteout name because its value doesn't change. */
		if (!name) {
			name = KMALLOC(dentry->d_name.len + 5, GFP_UNIONFS);
			if (!name) {
				err = -ENOMEM;
				goto out_free;
			}
			strcpy(name, ".wh.");
			strncat(name, dentry->d_name.name, dentry->d_name.len);
			name[4 + dentry->d_name.len] = '\0';
		}

		/* check if whiteout exists in this branch: lookup .wh.foo */
		wh_hidden_dentry = lookup_one_len(name, hidden_dir_dentry, 
			dentry->d_name.len + 4);
		if (IS_ERR(wh_hidden_dentry)) {
			err = PTR_ERR(wh_hidden_dentry);
			goto out_free;
		}

		if (wh_hidden_dentry->d_inode) {
			dput(wh_hidden_dentry);
			/* We found a whiteout so lets give up. */
			if (S_ISREG(wh_hidden_dentry->d_inode->i_mode)) {
				set_dbend(dentry, bindex);
				set_dbopaque(dentry, bindex);
				break;
			}
			err = -EIO;
			printk(KERN_NOTICE "EIO: Invalid whiteout entry type"
				" %d.\n", wh_hidden_dentry->d_inode->i_mode);
			goto out_free;
		}

		dput(wh_hidden_dentry);
		wh_hidden_dentry = NULL;

		/* Now do regular lookup; lookup foo */
		hidden_dentry = lookup_one_len(dentry->d_name.name, 
			hidden_dir_dentry, dentry->d_name.len);
		if (IS_ERR(hidden_dentry)) {
			err = PTR_ERR(hidden_dentry);
			goto out_free;
		}

		if (hidden_dentry->d_inode) {
			int opaque;

			/* number of positive dentries */
			dentry_count++;

			/* store underlying dentry */
			set_dtohd_index(dentry, bindex,  hidden_dentry);
			if (dbstart(dentry) == -1) {
				set_dbstart(dentry, bindex);
			}
			set_dbend(dentry, bindex);

			/* update parent directory's atime with the bindex */
			fist_copy_attr_atime(parent,
					hidden_dir_dentry->d_inode);

			/* If this is a directory, look for the opaque flag. */
			if (!S_ISDIR(hidden_dentry->d_inode->i_mode)) {
				if (dentry_count == 1);
					goto out_interpose;
				continue;
			}

			opaque = is_opaque_dir(dentry, bindex);
			if (opaque < 0) {
				err = opaque;
				goto out_free;
			}
			if (opaque) {
				set_dbend(dentry, bindex);
				set_dbopaque(dentry, bindex);
				break;
			}
		} else {
			/* Store the first negative dentry specially,
			 * because if they are all negative we need this 
			 * for future creates. */
			if (!is_first_negative && (dbstart(dentry) == -1)) {
				first_hidden_dentry = hidden_dentry;
				first_dentry_offset = bindex;
				is_first_negative = 1;
			} else {
				dput(hidden_dentry);
			}

			/* If we've only got negative dentries, then use the
			 * leftmost one. */
			if (is_first_negative && (bindex == bend) && (dentry_count == 0)) {
				set_dtohd_index(dentry, first_dentry_offset, first_hidden_dentry);
				set_dbstart(dentry, first_dentry_offset);
				set_dbend(dentry, first_dentry_offset);
				if (interpose_flag == INTERPOSE_REVAL) {
					if (dentry->d_inode) {
						PASSERT(itopd(dentry->d_inode));
						itopd(dentry->d_inode)->uii_stale = 1;
					}
				} else if (interpose_flag == INTERPOSE_REVAL_NEG) {
					/* We don't need to do anything. */
					ASSERT(dentry->d_inode == NULL);
				} else {
					d_add(dentry, NULL);
				}
				goto out;
			}
		} // else negative dentry
	} // end for

out_interpose:
	/* If we're holding onto the first negative dentry throw it out. */
	if (is_first_negative) {
		dput(first_hidden_dentry);
	}

	/* If we have found a positive dentry. */
	if (dentry_count) {
		err = unionfs_interpose(dentry, parent->i_sb, interpose_flag);
		if (err)
			goto out_drop;

		fist_checkinode(dentry->d_inode, 
					"unionfs_lookup OUT: dentry->d_inode:");
		fist_checkinode(parent, "unionfs_lookup OUT: dir");
	} else {
		if (interpose_flag == INTERPOSE_REVAL) {
			PASSERT(dentry);
			PASSERT(dentry->d_inode);
			PASSERT(itopd(dentry->d_inode));
			itopd(dentry->d_inode)->uii_stale = 1;
		} else {
			/* If no positive dentries were found, and we didn't 
			 * go to out because of the negative entry path, then
			 * we must have found a whiteout */
			hidden_dentry = lookup_one_len(dentry->d_name.name,
					hidden_dir_dentry, dentry->d_name.len);
			if (IS_ERR(hidden_dentry)) {
				err = PTR_ERR(hidden_dentry);
				goto out;
			}

			set_dtohd_index(dentry, bindex, hidden_dentry);
			set_dbstart(dentry, bindex);
			set_dbend(dentry, bindex);
			/* Don't hash our dentry if it is already hashed. */
			if (interpose_flag != INTERPOSE_REVAL_NEG)
				d_add(dentry, NULL);
		}
	}
	ASSERT(dbstart(dentry) >= 0);
	goto out;

out_drop:
	d_drop(dentry);

out_free:
	/* should dput all the underlying dentries on error condition */
	bstart = dbstart(dentry);
	if (bstart >= 0) {
		bend = dbend(dentry);
		for (bindex = bstart; bindex <= bend; bindex++) {
			hidden_dentry = dtohd_index(dentry, bindex);
			if (hidden_dentry)
				dput(hidden_dentry);
		}
	}
	KFREE(dtohd_ptr(dentry));
	dtohd_ptr(dentry) = NULL;

out_free_dtopd:
	KFREE(dtopd(dentry));
	dtopd_lhs(dentry) = NULL;

out:
	if (dtopd(dentry)) {
		ASSERT(dbend(dentry) <= dtopd(dentry)->udi_bcount);
		ASSERT(dbend(dentry) <= sbmax(dentry->d_sb));
		ASSERT(dbstart(dentry) >= 0);
	}
	if (name)
		KFREE(name);
	fist_print_dentry("OUT unionfs_lookup:", dentry);
	fist_print_inode("OUT unionfs_lookup:", parent);
	print_exit_status(err);
	return ERR_PTR(err);
}

/* lookup files/directories from index 'bstart' to the end of the 
 * super-block, or until an opaque directory.
 * Existing dentries are left alone.
 */
int unionfs_partial_lookup(struct dentry *dentry)
{
	int err = 0;
	int dentry_count = 0;
	int bstart, bindex, bopaque;
	const char *name;
	unsigned int namelen;
	struct dentry *hidden_dentry;
	struct dentry *hidden_dir_dentry;
	int positive;
	int opaque;

	print_entry_location();

	fist_print_dentry("IN: unionfs_partial_lookup", dentry);
	lock_dpd(dentry);
	bstart = dbstart(dentry);
	bopaque = dbopaque(dentry);

	name = dentry->d_name.name;
	namelen = dentry->d_name.len;

	positive = dentry->d_inode ? 1 :  0;

	for (bindex = bstart + 1; (bopaque == -1 || bindex <= bopaque)
			&& (bindex < sbmax(dentry->d_sb)); bindex++) {
		/* Check if there is already something there. */
		hidden_dentry = dtohd_index(dentry, bindex);
		if (hidden_dentry)
			continue;

		/* Find out hidden parent dentry */
		hidden_dir_dentry = dtohd_index(dentry->d_parent, bindex);

		/* continue if hidden parent doesn't exist */
		if (!hidden_dir_dentry)
			continue;

		/* lookup underlying file system */
		hidden_dentry = lookup_one_len(name, hidden_dir_dentry, namelen);
		if (IS_ERR(hidden_dentry)) {
			err = PTR_ERR(hidden_dentry);
			goto out;
		}

		if (!hidden_dentry->d_inode) {
			dput(hidden_dentry);
			continue;
		}
		/* this is not a negative dentry, so init dtohd with
		 * this new dentry.  Also, change bend.
		 */
		set_dtohd_index(dentry, bindex, hidden_dentry);
		set_dbend(dentry, bindex);
		dentry_count++;

		if (!S_ISDIR(hidden_dentry->d_inode->i_mode)) {
			if (dentry_count == 1) 
				goto out;
			continue;
		}

		opaque = is_opaque_dir(dentry, bindex);
		if (opaque < 0) {
			err = opaque;
			goto out;
		}
		if (opaque) {
			set_dbend(dentry, bindex);
			set_dbopaque(dentry, bindex);
			break;
		}
	} // end for

	if (dentry_count) {
		if (positive) {
			/* need to connect the new inodes to the pre-existing
			 * unionfs inode, unionfs_reinterpose
			 */
			unionfs_reinterpose(dentry);
		} else {
			int bstart;
			int bend;

			/* Somehow the file was created for us, so we need to
			 * interpose it with a new inode. */

			bstart = dbstart(dentry);
			bend = dbend(dentry);

			for (bindex = bstart; bindex <= bend; bindex++) {
				if (dtohd_index(dentry, bindex)) {
					if (dtohd_index(dentry, bindex)->d_inode) {
						set_dbstart(dentry, bindex);
						break;
					}
					dput(dtohd_index(dentry, bindex));
					set_dtohd_index(dentry, bindex, NULL);
				}
			}
			unionfs_interpose(dentry, dentry->d_sb, 0);
		}
	}

out:
	unlock_dpd(dentry);
	fist_print_dentry("OUT: unionfs_partial_lookup: ", dentry);
	print_exit_status(err);
	return err;
}

static int is_opaque_dir(struct dentry *dentry, int bindex) {
	int err = 0;
	struct dentry *hidden_dentry;
	struct dentry *wh_hidden_dentry;

	print_entry_location();

	if (!stopd(dentry->d_sb)->usi_diropaque)
		goto out;

	hidden_dentry = dtohd_index(dentry, bindex);

	PASSERT(hidden_dentry);
	PASSERT(hidden_dentry->d_inode);
	ASSERT(S_ISDIR(hidden_dentry->d_inode->i_mode));

	wh_hidden_dentry = lookup_one_len(UNIONFS_DIR_OPAQUE, hidden_dentry,
						sizeof(UNIONFS_DIR_OPAQUE) - 1);
	if (IS_ERR(wh_hidden_dentry)) {
		err = PTR_ERR(wh_hidden_dentry);
		goto out;
	}
	if (wh_hidden_dentry->d_inode)
		err = 1;
	dput(wh_hidden_dentry);
out:
	print_exit_status(err);
	return err;
}


/*
 * vim:shiftwidth=8
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
